#include <stdio.h>

#include "core/utils/utils.h"
#include "core/debug/evemon.h"
#include "core/debug/exception.h"
#include "core/message/message.h"
#include "core/types/type.h"

class msg_init : public message::base
{
public:
};

REGISTER_TYPENAME(msg_init);

class msg_update : public message::base
{
public:
};

REGISTER_TYPENAME(msg_update);

class msg_kill : public message::base
{
public:
};

REGISTER_TYPENAME(msg_kill);

/* observer sample class */
class observer : public message::handler<>
{
public:
	observer(void)
	{
		REGISTER_MESSAGE(observer, msg_init);
		REGISTER_MESSAGE(observer, msg_update);
		REGISTER_MESSAGE(observer, msg_kill);
	}

	HANDLE_MESSAGE(msg_init, msg)
	{
		_debug("observer::msg_init");
	}

	HANDLE_MESSAGE(msg_update, msg)
	{
		_debug("observer::msg_update");
	}

	HANDLE_MESSAGE(msg_kill, msg)
	{
		_debug("observer::msg_kill");
	}
};

// main program
int main(int argc, char** argv)
{
	try
	{
		observer obj;

		obj.handle(std::make_shared<msg_init>());
		obj.handle(std::make_shared<msg_update>());
		obj.handle(std::make_shared<msg_kill>());

		// return successfuly
		return 0;
	}
	catch (exception::base& e)
	{
		printf("error: %s\r\n", e.to_string().c_str());

		_critical("%s", e.to_string().c_str());
	}
	catch (std::exception& e)
	{
		printf("error: %s\r\n", e.what());

		_critical("%s", e.what());
	}
	catch (...)
	{
		printf("unknown error!\r\n");

		_critical("unknown error!");
	}

	// signal error
	return 1;
}